package nl.arttech.bouncing;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.RenderingHints;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.VolatileImage;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.Timer;

import java.awt.Image;

public class App extends JPanel implements 
ActionListener,
FocusListener,
MouseListener, 
MouseMotionListener {
	private static final long serialVersionUID = 1L;
	private static final int NUM_BALLS = 100;
	private static final int NUM_STRINGS = 7;
	private static final double ACCELERATION = 0.997;
	private static final double GRAVITY = 1.5;

	
	private VolatileImage contentBuffer;
	private boolean newOffscreenBufferNeeded;
	private int screenWidth = 640;
	private int screenHeight = 480;
	private int guiHeight = 40;
	
	private Vector<Ball> balls;
	private Vector<Impulse> impulses;	
	
	private double prevTime = 0;
	private double thisTime;
	private double dTime;
	private double minBoundX;
	private double maxBoundX;
	private double minBoundY;
	private double maxBoundY;
	
	private double mouseX;
	private double mouseY;
	private boolean mousePressed;
	private double prevMouseX;
	private double prevMouseY;
	private Vector<StringLine> strings;
	
	private String ipAddress = "127.0.0.1";
	private int port = 11015;
	private UdpServer server;
	
	private Toolkit tk = Toolkit.getDefaultToolkit();
	private Image ballImg;
	
	private JFrame frame;
	private JPanel contentPanel;    
	private JTextField ipField, portField;
    final Timer drawTimer = new Timer(20, this);

	public App() {
		super(true);
		
		addMouseListener(this);
		addMouseMotionListener(this);
		
		balls = new Vector<Ball>(0, 1);
		impulses = new Vector<Impulse>(0, 1);
		strings = new Vector<StringLine>(0, 1);
		
		minBoundX = screenWidth / 2. * -1. + 50.;
		maxBoundX = screenWidth / 2. - 50.;
		minBoundY = screenHeight / 2. * -1. + 50.;
		maxBoundY = screenHeight / 2. - 50.;
		
		makeBalls(NUM_BALLS);
		makeStrings(NUM_STRINGS);		
		
		server = new UdpServer(ipAddress, port);
		
		ballImg = tk.createImage("ball.png");
		
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            @Override
			public void run() {          	
                createAndShowGUI();
            }
        });
	}
	
	public void createAndShowGUI() {
		frame = new JFrame("OSC Bouncing Balls");
        frame.setBackground(new Color(0, 0, 0));
        frame.setLocation(50, 50);
        frame.setResizable(false);
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        
	    int spacingSize = 0;
	    contentPanel = new JPanel();
	    contentPanel.setLayout(new BoxLayout(contentPanel, BoxLayout.Y_AXIS));
	    contentPanel.setBorder(BorderFactory.createEmptyBorder(spacingSize,spacingSize,spacingSize, spacingSize));
	    contentPanel.setBackground(new Color(0, 0, 0));
	    
        this.setPreferredSize(new Dimension(screenWidth, screenHeight));
	    contentPanel.add(this);
	    
	    JPanel guiPanel = getGuiPanel();
	    guiPanel.setPreferredSize(new Dimension(screenWidth, guiHeight));
	    contentPanel.add(guiPanel);
	    
	    Container contentPane = frame.getContentPane();
	    contentPane.add(contentPanel);   

		frame.pack();
        frame.setVisible(true);
        
		prevTime = System.currentTimeMillis();
		
        drawTimer.start();          
	}
	
	private JPanel getGuiPanel() {
		JPanel guiPanel = new JPanel();
		guiPanel.setLayout(new GridBagLayout());
		guiPanel.setBackground(new Color(255, 255, 255));
		GridBagConstraints c = new GridBagConstraints();	
		
		JLabel ipLabel = new JLabel("Destination IP and port");
		c.fill = GridBagConstraints.NONE;
		c.gridx = 0;
		c.gridy = 0;
		c.gridwidth = 1;
		c.anchor = GridBagConstraints.LINE_START;
		c.insets = new Insets(0, 0, 0, 10);
		guiPanel.add(ipLabel, c);			
		
		ipField = new JTextField(ipAddress);
		ipField.setColumns(8);
		ipField.addActionListener(this);
		ipField.addFocusListener(this);
        c.fill = GridBagConstraints.NONE;
        c.gridx = 1;
        c.gridy = 0;
        c.gridwidth = 1;
		c.anchor = GridBagConstraints.LINE_START;
		c.insets = new Insets(0, 0, 0, 10);
        guiPanel.add(ipField, c);	
		
		portField = new JTextField("" + port);
		portField.setColumns(5);
		portField.addActionListener(this);
		portField.addFocusListener(this);
        c.fill = GridBagConstraints.NONE;
        c.gridx = 2;
        c.gridy = 0;
        c.gridwidth = 1;
		c.anchor = GridBagConstraints.LINE_START;
		c.insets = new Insets(0, 0, 0, 0);
        guiPanel.add(portField, c);	        
        
		return guiPanel;
	}	
	
	private void process() {
		Vector<Impulse> impulsesToRemove = new Vector<Impulse>(0, 1);
		for (Impulse impulse : impulses) {
			if (impulse.getPower(thisTime) < 0.01) impulsesToRemove.add(impulse);
		}
		
		for (Impulse impulse : impulsesToRemove) {
			impulses.remove(impulse);
		}
		
		for (Ball ball : balls) {
			for (Impulse impulse : impulses) {	
				double distX = impulse.x - ball.x;
				double distY = ball.y - impulse.y;
	
				double totalDist = Math.sqrt(distX * distX + distY * distY); 
				double angle = distY > 0 ? Math.acos(distX / totalDist) : Math.acos(distX / totalDist) * -1.;
							
				if (totalDist < impulse.getRange(thisTime)) {
					ball.speedX -= impulse.getPower(thisTime) * Math.cos(angle);
					ball.speedY += impulse.getPower(thisTime) * Math.sin(angle);
				}
			}

			ball.speedX *= ACCELERATION;
			ball.speedY *= ACCELERATION;

			ball.speedY += GRAVITY;
			
			double nextX = ball.x + dTime * ball.speedX;
			double nextY = ball.y + dTime * ball.speedY;
			
			if (nextX < minBoundX || nextX > maxBoundX) { 
				ball.speedX *= -1.;
			}
			if (nextY < minBoundY || nextY > maxBoundY) {
				ball.speedY *= -1.;
			}
			//double prevBallX = ball.x;
			double prevBallY = ball.y;
			
			ball.x += dTime * ball.speedX;
			ball.y += dTime * ball.speedY;
			
			for (StringLine string : strings) {
				string.checkHits(thisTime);
				if (ball.y > string.pos && prevBallY < string.pos || ball.y < string.pos && prevBallY > string.pos) {
					string.addHit(System.currentTimeMillis(), ball.speedY, ball.x);
					server.sendHitPacket("/string", strings.indexOf(string), ball.speedY, ball.x/this.getWidth());
				}
			}
		}
	}
	
    @Override
    protected void paintComponent(Graphics g) {
		checkBuffer();
		
		if (mousePressed) {
			if (mouseX != prevMouseX || mouseY != prevMouseY) {
				impulses.add(new Impulse(mouseX, mouseY, System.currentTimeMillis()));	
			}
		}
		prevMouseX = mouseX;
		prevMouseY = mouseY;
		
		synchronized(this) {
			thisTime = System.currentTimeMillis();
			dTime = (thisTime - prevTime) / 1000.;
			process();
			renderContent(contentBuffer.createGraphics());
			prevTime = thisTime;
		}

		g.drawImage(contentBuffer, 0, 0, this); 		// draw end result buffer to out graphics
    }	
	
	private void renderContent(Graphics2D g) {	
		g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);		
		
		g.clearRect(0, 0, screenWidth, screenHeight);
		
		g.setColor(Color.white);
		g.drawRect((int)(minBoundX + screenWidth / 2.), (int)(minBoundY + screenHeight / 2.), (int)(maxBoundX - minBoundX), (int)(maxBoundY - minBoundY));
		int left = (int)(minBoundX + screenWidth / 2.);
		int right = (int)(maxBoundX + screenWidth / 2.);
		for (StringLine string : strings) {
			double totalForce = string.getTotalForcePercentage(thisTime) / (NUM_BALLS / NUM_STRINGS);
			
			int colorValue = Math.min((int)(totalForce * 255), 100);
			int midColorValue = Math.min(colorValue + 80, 255);
			double top = string.pos + screenHeight / 2.;
			
			g.setColor(new Color(midColorValue, midColorValue, midColorValue));
			g.drawLine(left, (int)top, right, (int)top);
			g.setColor(new Color(colorValue, colorValue, colorValue));
			g.drawLine(left, (int)top - 1, right, (int)top - 1);
			g.drawLine(left, (int)top + 1, right, (int)top + 1);		
		}
		
		for (Impulse impulse : impulses) {
			g.setColor(new Color(255, 255, 255, (int)(Math.min(impulse.getPower(thisTime) * 128, 255))));
			drawCircle(g, impulse.x, impulse.y, impulse.getRange(thisTime) * 2);
		}
		
		for (Ball thisBall : balls) {
			g.setColor(thisBall.color);
			fillCircle(g, thisBall.x, thisBall.y, 20);
		}
	}
	
	public void drawCircle(Graphics2D g, double x, double y, double r) {
		g.drawOval((int)(x + screenWidth / 2. - r / 2.), (int)(y + screenHeight / 2. - r /2.), (int)r, (int)r);
	}

	public void fillCircle(Graphics2D g, double x, double y, double r) {
		g.fillOval((int)(x + screenWidth / 2. - r / 2.), (int)(y + screenHeight / 2. - r /2.), (int)r, (int)r);
		g.drawImage(ballImg, (int)(x + screenWidth / 2. - r / 2.), (int)(y + screenHeight / 2. - r /2.), null);
	}	
	
	public void drawCoords(Graphics2D g, double x, double y, double speedX, double speedY) {
		speedX = (int)(speedX * 100) / 100.;
		speedY = (int)(speedY * 100) / 100.;		
		g.drawString("(" + speedX + ", " + speedY +")", (int)(x + screenWidth / 2.), (int)(y + screenHeight / 2.));
	}
	
	private void makeBalls(int amount) {
		for (int i = 0; i < amount; i++) {
			balls.add(new Ball(Math.random() * (screenWidth - 100) - screenWidth / 2. + 50, Math.random() * (screenHeight - 100) - screenHeight / 2 + 50, 200));
		}
	}
	
	private void makeStrings(int amount) {
		for (double i = 1; i < amount; i++) {
			double top = i * (maxBoundY - minBoundY) / amount - (maxBoundY - minBoundY) / 2.;
			strings.add(new StringLine(top));
		}
	}

	public void checkBuffer() {
		if (contentBuffer == null) {
			contentBuffer = createVolatileImage(screenWidth,screenHeight);
			newOffscreenBufferNeeded = true;
		}
		else if (contentBuffer.contentsLost()) {
			if (contentBuffer.validate(getGraphicsConfiguration()) == VolatileImage.IMAGE_INCOMPATIBLE) {
				newOffscreenBufferNeeded = true;
			}
		}
		
		if (newOffscreenBufferNeeded) {
			System.out.println("building new offscreen draw buffer");
			contentBuffer = createVolatileImage(screenWidth,screenHeight);
			newOffscreenBufferNeeded = false;
		}		
	}	

	public static void main(String[] args) {
		new App();
	}

	public void mouseReleased(MouseEvent e) {	
		mousePressed = false;
		double impulseX = e.getX() - screenWidth / 2.;
		double impulseY = e.getY() - screenHeight / 2.;
		double impulseTime = System.currentTimeMillis();
		
		synchronized(this) {
			impulses.add(new Impulse(impulseX, impulseY, impulseTime));
		}
//		post("mouse: " + impulseX + ", " + impulseY);
	}
	
	public void mouseClicked(MouseEvent e) {}

	public void mouseEntered(MouseEvent e) {}

	public void mouseExited(MouseEvent e) {}

	public void mousePressed(MouseEvent e) {
		mousePressed = true;
	}

	public void mouseDragged(MouseEvent e) {
		mouseX = e.getX() - screenWidth / 2.;
		mouseY = e.getY() - screenHeight / 2.;		
	}

	public void mouseMoved(MouseEvent e) {}

    @Override
    public void actionPerformed(ActionEvent e) {
    	// drawTimer leads here
    	if (e.getSource().equals(drawTimer)) {
    		this.repaint();
    	}
    	else {
	    	if (e.getSource().equals(ipField) || e.getSource().equals(portField)) {
	    		updateServerFromFields();
	    	}
    	}
    }

	@Override
	public void focusGained(FocusEvent e) {}

	@Override
	public void focusLost(FocusEvent e) {
    	if (e.getSource().equals(ipField) || e.getSource().equals(portField)) {
    		updateServerFromFields();
    	}
	}
	
	private void updateServerFromFields() {
		if (ipAddress.equals(ipField.getText()) && ("" + port).equals(portField.getText())) {
			return;
		}
		else {
			ipAddress = ipField.getText();	
			try {
				port = Integer.decode(portField.getText());
			}
			catch(NumberFormatException ex) {
				portField.setText("" + port);
			}
			server.close();
			server = new UdpServer(ipAddress, port);
			
			Utils.post("Server updated, now sending to " + ipAddress + ", " + port);
		}
	}	
}
