package nl.arttech.bouncing;
import java.io.*;
import java.net.*;


/**
 * OpenSoundControl UDP Server.
 *
 * Based on UdpServer by Ben Chun - ben@benchun.net
 *
 * @author Mattijs Kneppers - mattijs@arttech.nl
 */

public class UdpServer
{
    DatagramSocket oscSendSocket;
	InetAddress local = null;
    static boolean debugUdp = false;
    int port;
    InetAddress address;

    public UdpServer(String ip, int port) {
    	this.port = port;
    	try {
			address = InetAddress.getByName(ip);
		} catch (UnknownHostException e1) {
			e1.printStackTrace();
		} 
		
		try {
			local = InetAddress.getLocalHost();
			createServer();
		}
		catch (UnknownHostException e) {
			System.out.println("Error, couldn't resolve local host");
		}
    }
    
	public void createServer() {
		if (local != null) {
			try {
				oscSendSocket = new DatagramSocket();
			}
			catch( SocketException se ) { se.printStackTrace();}
					
			if (debugUdp)  {
				if (oscSendSocket != null) System.out.println("UDP Sending OSC socket started. Address and port of client are used for sending.");
			}
		}
	}
	
	public void sendHitPacket(String address, int string, double force, double location) {
//		post("sending packet: " + snaar + ", " + force + ", " + location);
		OscMessage msg = new OscMessage(address);
		msg.addArg(new Character('i'), new Integer(string));
		msg.addArg(new Character('f'), new Float(force));
		msg.addArg(new Character('f'), new Float(location));
		OscPacket packet = new OscPacket();
		packet.addMessage(msg);
		sendPacket(packet);
	}

    public void sendPacket(OscPacket oscPacket)
    {
		oscPacket.address = address;
		oscPacket.port = port;

		byte[] packetBytes = {0};
		try {
			packetBytes = oscPacket.getByteArray();
		} catch (IOException e) {
			e.printStackTrace();
		}

		DatagramPacket packet = new DatagramPacket(packetBytes, packetBytes.length, oscPacket.address, oscPacket.port);
		if (debugUdp) Utils.post("___Sending UDP packet___"); 
		if (debugUdp) printBytes(packetBytes);
		if (debugUdp) Utils.post(oscPacket.toString());		
		
		try {
			oscSendSocket.send(packet);
		} catch (IOException e) {
			Utils.post("sending packet failed");
			e.printStackTrace();
		}
    }
    public void killServer()
    {
		if (oscSendSocket.isClosed())
		{
			System.out.println("Error stopping UDP server: UDP server was already stopped");
		}
		else
		{
			oscSendSocket.close();
			if (debugUdp) System.out.println("UDP server stopped");
		}
    }
    public static void printBytes(byte[] byteArray)
    {
    	char toPrint =("*").charAt(0);
		for (int i=0; i<byteArray.length; i++)
		{
			if (byteArray[i] != 0) toPrint = (char)byteArray[i];
			else toPrint = ("*").charAt(0);
			System.out.print(byteArray[i] + " (" + toPrint + ")  ");
			if ((i+1)%4 == 0)
			System.out.print("\n");
		}
    }
	public void close()
	{
		killServer();
	} 
}